from os import environ
from os.path import join, isfile, basename, dirname
from time import sleep
from multiprocessing.pool import Pool, get_context
from math import ceil
import gc
from silx.io import get_data
from silx.io.url import DataUrl
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase
from ... import version as nabu_version
from ...utils import check_supported, first_generator_item, subdivide_into_overlapping_segment, split_list
from ...resources.logger import LoggerOrPrint, PrinterLogger, Logger
from ...resources.utils import is_hdf5_extension, get_available_ram_GB
from ...io.writer import merge_hdf5_files, NXProcessWriter
from ...preproc.phase import compute_paganin_margin
from ...processing.histogram import PartialHistogram, add_last_bin, hist_as_2Darray
from .chunked import ChunkedPipeline
from .computations import estimate_max_chunk_size
from ..tasks import FakeTaskProcessor, TaskProcessor


def variable_idxlen_sort(fname):
    return int(fname.split("_")[-1].split(".")[0])


def _proc_collect_cuda_gpus(x):
    from nabu.cuda.utils import collect_cuda_gpus

    return collect_cuda_gpus()


def _collect_cuda_gpus(safe=True):
    if not safe:
        return _proc_collect_cuda_gpus(None)
    try:
        ctx = get_context("spawn")
        with Pool(1, context=ctx) as p:
            v = p.map(_proc_collect_cuda_gpus, [1])[0]
    except AssertionError:
        # Can get AssertionError: daemonic processes are not allowed to have children
        # if the calling code is already a subprocess
        return _proc_collect_cuda_gpus(None)
    return v


class NannyPipeline:
    """
    A "nanny" pipeline that runs in a separate process
    """

    def __init__(
        self, process_config, backend="numpy", pipeline_mode="chunked", logging_options=None, **pipeline_kwargs
    ):
        self.process_config = process_config
        self.previous_task = None
        self.pipeline = None
        self.worker_id = 0  # placeholder
        self._get_pipeline_class(backend, pipeline_mode)
        self.pipeline_kwargs = pipeline_kwargs
        self._create_logger(logging_options)

    def _create_logger(self, logging_options):
        if logging_options is None:
            self.logger = PrinterLogger()
            return
        my_logging_options = logging_options.copy()
        logname = logging_options.pop("logname", "nabu")
        self.logger = Logger(logname, **my_logging_options)

    def _get_pipeline_class(self, backend, pipeline_mode):
        self.backend = backend
        # Check were done before
        if backend == "cuda":
            # Don't initialize cuda before creating a new process (with fork) !
            from .chunked_cuda import CudaChunkedPipeline

            self._pipeline_cls = CudaChunkedPipeline
        else:
            self._pipeline_cls = ChunkedPipeline
        self._pipeline_mode = pipeline_mode

    def _get_chunk_length(self, task):
        # normally not supported
        if self._pipeline_mode == "helical":
            (start_z, end_z) = task["sub_region"]
            return end_z - start_z
        #
        (start_angle, end_angle), (start_z, end_z), _ = task["sub_region"]
        if self._pipeline_mode == "chunked":
            return end_z - start_z
        else:
            return end_angle - start_angle

    def _instantiate_pipeline(self, task):
        self.logger.set_prefix("  " * (self.worker_id) + f"[{self.worker_id}] ")
        self.logger.debug("Creating a new pipeline object")
        chunk_shape = tuple(s[1] - s[0] for s in task["sub_region"])
        args = [self.process_config, chunk_shape]
        pipeline_init_kwargs = self.pipeline_kwargs.copy()
        if self.backend == "cuda":
            pipeline_init_kwargs["cuda_options"] = {"device_id": self.worker_id}
        pipeline_init_kwargs["use_grouped_mode"] = self._pipeline_mode == "grouped"
        pipeline = self._pipeline_cls(*args, logger=self.logger, margin=task["margin"], **pipeline_init_kwargs)
        self.pipeline = pipeline

    def _instantiate_pipeline_if_necessary(self, current_task, other_task):
        """
        Instantiate a pipeline only if current_task has a different "delta z" than other_task
        """
        if self.pipeline is None:
            self._instantiate_pipeline(current_task)
            return
        length_cur = self._get_chunk_length(current_task)
        length_other = self._get_chunk_length(other_task)
        if length_cur != length_other or current_task["margin"] != other_task["margin"]:
            self.logger.debug("Destroying pipeline instance and releasing memory")
            self._destroy_pipeline()
            self._instantiate_pipeline(current_task)

    def _destroy_pipeline(self):
        self.pipeline = None

        gc.collect()
        # Not elegant, but for now the only way to release Cuda memory
        import cupy

        cupy.get_default_memory_pool().free_all_blocks()
        cupy.get_default_pinned_memory_pool().free_all_blocks()
        gc.collect()

    def process_task(self, task):
        """
        'task' is in the form {"sub_region": ((a0, a1), (z0, z1), (x0, x1)), "margin": (m_v, m_h)}
        """
        prev_task = self.previous_task
        self.logger.info("Processing sub-volume %s" % (str(task["sub_region"])))
        self._instantiate_pipeline_if_necessary(task, prev_task)
        self.pipeline.process_chunk(task["sub_region"], output_start_z=task.get("output_start_z", None))

        if self._pipeline_mode == "grouped":
            task_result = self.pipeline.datadump_manager.data_dump["sinogram"].fname
        else:
            task_result = self.pipeline.writer.fname if self.pipeline.writer is not None else None

        task_result = {
            "sub_region": task["sub_region"],
            "result": task_result,
            "histogram": None,
            "data_dumps": {},
        }
        if self.pipeline.writer is not None and self.pipeline.writer.histogram:
            task_result["histogram"] = self.pipeline.writer.histogram_writer.fname
        if len(self.pipeline.datadump_manager.data_dump) > 0:
            for step_name, writer in self.pipeline.datadump_manager.data_dump.items():
                task_result["data_dumps"][step_name] = writer.fname

        self.previous_task = task
        return task_result


class FullFieldReconstructor:
    """
    A reconstructor spawns and manages Pipeline objects, depending on the current chunk/group size.
    """

    _available_backends = ["cuda", "numpy"]

    _process_name = "reconstruction"

    default_advanced_options = {
        "gpu_mem_fraction": 0.9,
        "cpu_mem_fraction": 0.9,
        "chunk_size": None,
        "margin": None,
        "force_grouped_mode": False,
        "disable_multiprocessing_if_possible": True,
    }

    def __init__(self, process_config, logger=None, backend="cuda", extra_options=None, cuda_options=None):
        """
        Initialize a Reconstructor object.
        This class is used for managing pipelines.

        Parameters
        ----------
        process_config: ProcessConfig object
            Data structure with process configuration
        logger: Logger, optional
            logging object
        backend: str, optional
            Which backend to use. Available are: "cuda", "numpy".
        extra_options: dict, optional
            Dictionary with advanced options. Please see 'Other parameters' below
        cuda_options: dict, optional
            Dictionary with cuda options passed to `nabu.cuda.processing.CudaProcessing`


        Other Parameters
        -----------------
        Advanced options can be passed in the 'extra_options' dictionary. These can be:

           - cpu_mem_fraction: 0.9,
           - gpu_mem_fraction: 0.9,
           - chunk_size: None,
           - margin: None,
           - force_grouped_mode: False
        """
        self.logger = LoggerOrPrint(logger)
        self.process_config = process_config
        self._set_extra_options(extra_options)
        self._get_reconstruction_range()
        self._get_resources()
        self._get_backend(backend, cuda_options)
        self._compute_margin()
        self._compute_max_chunk_size()
        self._get_pipeline_mode()
        self._build_tasks()

        self._do_histograms = self.process_config.nabu_config["postproc"]["output_histogram"]
        self._reconstruction_output_format_is_hdf5 = is_hdf5_extension(
            self.process_config.nabu_config["output"]["file_format"]
        )
        self._histogram_merged = False
        self._histogram_merged = False

    #
    # Initialization
    #

    def _set_extra_options(self, extra_options):
        self.extra_options = self.default_advanced_options.copy()
        self.extra_options.update(extra_options or {})
        self.gpu_mem_fraction = self.extra_options["gpu_mem_fraction"]
        self.cpu_mem_fraction = self.extra_options["cpu_mem_fraction"]

    def _get_reconstruction_range(self):
        rec_region = self.process_config.rec_region  # without binning
        self.z_min = rec_region["start_z"]
        # In the user configuration, the upper bound is included: [start_z, end_z].
        # In python syntax, the upper bound is not: [start_z, end_z[
        self.z_max = rec_region["end_z"] + 1
        self.delta_z = self.z_max - self.z_min
        # Cache some volume info
        self.n_angles = self.process_config.n_angles(subsampling=False)
        self.n_z, self.n_x = self.process_config.radio_shape(binning=False)

    def _get_resources(self):
        self.resources = {}
        self._get_gpu()
        self._get_memory()

    def _get_memory(self):
        self.resources["mem_avail_GB"] = get_available_ram_GB()
        self.cpu_mem = self.resources["mem_avail_GB"] * self.cpu_mem_fraction

    def _get_gpu(self):
        resources_config = self.process_config.nabu_config.get("resources", {})

        # Number of workers
        self.n_workers = resources_config.get("workers", 1)
        n_gpus = resources_config.get("gpus", 1)
        gpu_ids = resources_config.get("gpu_id", [])
        if gpu_ids != [] and len(gpu_ids) != n_gpus:
            self.logger.warning(
                "Found a discrepancy between 'gpus' and 'gpu_id' in config file. The parameter 'gpu_id' takes precedence."
            )
            n_gpus = len(gpu_ids)
        if gpu_ids == []:
            gpu_ids = list(range(n_gpus))  # list(avail_gpus.keys())[:n_gpus] ?

        self._use_multiprocessing = True
        if (self.n_workers == 1 or n_gpus == 1) and self.extra_options.get("disable_multiprocessing_if_possible", True):
            self.logger.debug("Disabling multi-processing for single-worker reconstruction")
            self._use_multiprocessing = False

        # GPUS / gpu id
        avail_gpus = _collect_cuda_gpus(safe=not (self.extra_options.get("disable_multiprocessing", False))) or {}
        self.resources["gpus"] = avail_gpus
        if len(avail_gpus) == 0:
            self._has_cupy = False
            return
        self._has_cupy = True

        if len(gpu_ids) > len(avail_gpus):
            self.logger.error(f"Requested {len(gpu_ids)} GPUs which is more than the {len(avail_gpus)} available GPUs")
            gpu_ids = gpu_ids[: len(avail_gpus)]
        self.resources["gpu_id"] = gpu_ids

        if len(gpu_ids) != self.n_workers:
            self.logger.warning(
                f"Requested n_workers={self.n_workers} and gpu_id={gpu_ids}. For now nabu can only use one worker per gpu. Resetting n_workers={len(gpu_ids)}"
            )
            self.n_workers = len(gpu_ids)

    def _get_backend(self, backend, cuda_options):
        # self._pipeline_cls = ChunkedPipeline
        check_supported(backend, self._available_backends, "backend")
        if backend == "cuda":
            self.cuda_options = cuda_options
            if len(self.resources["gpus"]) == 0:
                # Not sure if an error should be raised in this case
                self.logger.error("Could not find any cuda device. Falling back on numpy/CPU backend.")
                backend = "numpy"
            else:
                for gpu_id in self.resources["gpu_id"]:
                    gpu_desc = self.resources["gpus"][gpu_id]
                    gpu_mem_tot = gpu_desc["memory_total_GB"]
                    gpu_mem_avail = gpu_desc["memory_avail_GB"]
                    if gpu_mem_avail / gpu_mem_tot < 0.5:
                        self.logger.warning(
                            f"It looks like the target GPU number {gpu_id} ({gpu_desc['name']}) is in use (less than half memory available)"
                        )
                # Assume all GPU have the same amount of memory # FIXME
                self.gpu_mem = gpu_mem_avail * self.gpu_mem_fraction
        if backend == "cuda" and not (self._has_cupy):
            raise RuntimeError("cupy is not avilable")

        # In the special case where we use astra for cone-beam (FDK) reconstruction,
        # astra can do the data split automatically better than nabu
        # In this case, do all the preprocessing CPU, and only FDK on GPU
        rec_config = self.process_config.nabu_config["reconstruction"]
        if rec_config["method"] == "cone" and rec_config.get("implementation") == "astra":
            self.logger.info(
                "Using reconstruction method='cone' and implementation='astra'. All pre-processing will be done on CPU, and final reconstruction with astra"
            )
            backend = "numpy"
        # ---

        self.backend = backend

    def _compute_max_chunk_size(self):
        """
        Compute the maximum number of (partial) radios that can be processed in memory.
        Ideally, the processing is done by reading N lines of all the projections.
        This function estimates max_chunk_size = N_max, the maximum number of lines that can be read
        in all the projections while still fitting the memory.
        On the other hand, if a "processing margin" is needed (eg. phase retrieval), then we need to read
        at least N_min = 2 * margin_v + n_slices lines of each image.
        For large datasets, we have N_min > max_chunk_size, so we can't read lines from all the projections.
        """
        user_chunk_size = self.extra_options["chunk_size"]
        if user_chunk_size is not None:
            self.chunk_size = user_chunk_size
        else:
            self.cpu_max_chunk_size = estimate_max_chunk_size(
                self.cpu_mem, self.process_config, pipeline_part="all", step=5
            )
            self.chunk_size = self.cpu_max_chunk_size
            if self.backend == "cuda":
                self.gpu_max_chunk_size = estimate_max_chunk_size(
                    self.gpu_mem,
                    self.process_config,
                    pipeline_part="all",
                    step=5,
                )
                self.chunk_size = min(self.gpu_max_chunk_size, self.cpu_max_chunk_size)
        self.chunk_size = min(self.chunk_size, self.n_z)

    def _compute_max_group_size(self):
        """
        Compute the maximum number of (partial) images that can be processed in memory
        """

        #
        # "Group size" (i.e, how many radios can be processed in one pass for the first part of the pipeline)
        #
        user_group_size = self.extra_options.get("chunk_size", None)
        if user_group_size is not None:
            self.group_size = user_group_size
        else:
            self.cpu_max_group_size = estimate_max_chunk_size(
                self.cpu_mem,
                self.process_config,
                pipeline_part="radios",
                n_rows=min(2 * self._margin_v + self.delta_z, self.n_z),
                step=10,
            )
            self.group_size = self.cpu_max_group_size
            if self.backend == "cuda":
                self.gpu_max_group_size = estimate_max_chunk_size(
                    self.gpu_mem,
                    self.process_config,
                    pipeline_part="radios",
                    n_rows=min(2 * self._margin_v + self.delta_z, self.n_z),
                    step=10,
                )
                self.group_size = min(self.gpu_max_group_size, self.cpu_max_group_size)
        self.group_size = min(self.group_size, self.n_angles)

        #
        # "sinos chunk size" (i.e, how many sinograms can be processed in one pass for the second part of the pipeline)
        #
        self.cpu_max_chunk_size_sinos = estimate_max_chunk_size(
            self.cpu_mem,
            self.process_config,
            pipeline_part="sinos",
            step=10,
        )
        self.chunk_size_sinos = self.cpu_max_chunk_size_sinos
        if self.backend == "cuda":
            self.gpu_max_chunk_size_sinos = estimate_max_chunk_size(
                self.gpu_mem,
                self.process_config,
                pipeline_part="sinos",
                step=5,
            )
            self.chunk_size_sinos = min(self.gpu_max_chunk_size_sinos, self.cpu_max_chunk_size_sinos)
        self.chunk_size_sinos = min(self.chunk_size_sinos, self.delta_z)

    def _get_pipeline_mode(self):
        # "Pipeline mode" means we either process data chunks of type [:, delta_z, :] or [delta_theta, :, :].
        # The first case is better in terms of performances and should be preferred.
        # However, in some cases, it's not possible to use it (eg. high "delta_z" because of margin)
        chunk_size_for_one_slice = 1 + 2 * self._margin_v  # TODO ignore margin when resuming from sinogram ?
        chunk_is_too_small = False
        if chunk_size_for_one_slice > self.chunk_size:
            msg = str(
                "Margin is %d, so we need to process at least %d detector rows. However, the available memory enables to process only %d rows at once"
                % (self._margin_v, chunk_size_for_one_slice, self.chunk_size)
            )
            chunk_is_too_small = True
        if self._margin_v > self.chunk_size // 3:
            n_slices = max(1, self.chunk_size - (2 * self._margin_v))
            n_stages = ceil(self.delta_z / n_slices)
            if n_stages > 1:
                msg = str("Margin (%d) is too big for chunk size (%d)" % (self._margin_v, self.chunk_size))
                chunk_is_too_small = True
        force_grouped_mode = self.extra_options.get("force_grouped_mode", False)
        if force_grouped_mode:
            msg = "Forcing grouped mode"
        if self.process_config.processing_options.get("phase", {}).get("method", None) == "CTF":
            force_grouped_mode = True
            msg = "CTF phase retrieval needs to process full radios"
        if (self.process_config.dataset_info.detector_tilt or 0) > 15:
            force_grouped_mode = True
            msg = "Radios rotation with a large angle needs to process full radios"
        if self.process_config.processing_options.get("flatfield", {}).get("method", "default") == "pca":
            force_grouped_mode = True
            msg = "PCA-Flatfield normalization needs to process full radios"
        if self.process_config.resume_from_step == "sinogram" and force_grouped_mode:
            self.logger.warning("Cannot use grouped-radios processing when resuming from sinogram")
            force_grouped_mode = False
        if not (chunk_is_too_small or force_grouped_mode):
            # Default case (preferred)
            self._pipeline_mode = "chunked"
            self.chunk_shape = (self.n_angles, self.chunk_size, self.n_x)
        else:
            # Fall-back mode (slower)
            self.logger.warning(msg)  # pylint: disable=E0606
            self._pipeline_mode = "grouped"
            self._compute_max_group_size()
            self.chunk_shape = (self.group_size, self.delta_z, self.n_x)
            self.logger.info("Using 'grouped' pipeline mode")

    #
    # "Margin"
    #

    def _compute_margin(self):
        user_margin = self.extra_options.get("margin", None) or self.process_config.nabu_config["pipeline"].get(
            "processing_margin", None
        )
        if self.process_config.resume_from_step == "sinogram":
            self.logger.debug("Margin not needed when resuming from sinogram")
            margin_v, margin_h = 0, 0
        elif user_margin is not None and user_margin > 0:
            margin_v, margin_h = user_margin, user_margin
            self.logger.info("Using user-defined margin: %d" % user_margin)
        else:
            unsharp_margin = self._compute_unsharp_margin()
            phase_margin = self._compute_phase_margin()
            translations_margin = self._compute_translations_margin()
            cone_margin = self._compute_cone_overlap()
            rot_margin = self._compute_rotation_margin()
            margin_v = max(unsharp_margin[0], phase_margin[0], translations_margin[0], cone_margin[0], rot_margin[0])
            margin_h = max(unsharp_margin[1], phase_margin[1], translations_margin[1], cone_margin[1], rot_margin[1])
            if margin_v > 0:
                self.logger.info("Estimated margin: %d pixels" % margin_v)

        margin_v = min(margin_v, self.n_z)
        margin_h = min(margin_h, self.n_x)
        self._margin = margin_v, margin_h
        self._margin_v = margin_v

    def _compute_unsharp_margin(self):
        if "unsharp_mask" not in self.process_config.processing_steps:
            return (0, 0)
        opts = self.process_config.processing_options["unsharp_mask"]
        sigma = opts["unsharp_sigma"]
        # nabu uses cutoff = 4
        cutoff = 4
        gaussian_kernel_size = ceil(2 * cutoff * sigma + 1)
        self.logger.debug("Unsharp mask margin: %d pixels" % gaussian_kernel_size)
        return (gaussian_kernel_size, gaussian_kernel_size)

    def _compute_phase_margin(self):
        phase_options = self.process_config.processing_options.get("phase", None)
        if phase_options is None:
            margin_v, margin_h = (0, 0)
        elif phase_options["method"] == "paganin":
            radio_shape = self.process_config.dataset_info.radio_dims[::-1]
            margin_v, margin_h = compute_paganin_margin(
                radio_shape,
                distance=phase_options["distance_m"],
                energy=phase_options["energy_kev"],
                delta_beta=phase_options["delta_beta"],
                pixel_size=phase_options["pixel_size_m"],
                padding=phase_options["padding_type"],
            )
        elif phase_options["method"] == "CTF":
            # The whole projection has to be processed!
            margin_v = ceil(
                (self.n_z - self.delta_z) / 2
            )  # not so elegant. Use a dedicated flag eg. _process_whole_image ?
            margin_h = 0  # unused for now
        else:
            margin_v, margin_h = (0, 0)
        return (margin_v, margin_h)

    def _compute_translations_margin(self):
        translations = self.process_config.dataset_info.translations
        if translations is None:
            return (0, 0)
        margin_h_v = []
        for i in range(2):
            transl = translations[:, i]
            margin_h_v.append(ceil(max([transl.max(), (-transl).max()])))
        self.logger.debug("Maximum vertical displacement: %d pixels" % margin_h_v[1])
        return tuple(margin_h_v[::-1])

    def _compute_cone_overlap(self):
        rec_cfg = self.process_config.processing_options.get("reconstruction", {})
        rec_method = rec_cfg.get("method", None)
        if rec_method != "cone":
            return (0, 0)
        d1 = rec_cfg["source_sample_dist"]
        d2 = rec_cfg["sample_detector_dist"]
        n_z, _ = self.process_config.radio_shape(binning=True)

        # delta_z = self.process_config.rec_delta_z  # accounts_for_binning
        # overlap = ceil(delta_z * d2 / (d1 + d2))  # sqrt(2) missing ?

        max_overlap = ceil(n_z * d2 / (d1 + d2))  # sqrt(2) missing ?
        max_overlap = max(max_overlap, 10)  # use at least 10 pixels

        return (max_overlap, 0)

    def _compute_rotation_margin(self):
        if "tilt_correction" in self.process_config.processing_steps:
            # Partial radios rotation yields too much error in single-slice mode
            # Forcing a big margin circumvents the problem
            # This is likely to trigger the 'grouped mode', but perhaps grouped mode should always be used when rotating radios
            nz, nx = self.process_config.radio_shape(binning=True)
            return nz // 3, nx // 3
        else:
            return 0, 0

    def _ensure_good_chunk_size_and_margin(self):
        """
        Check that "chunk_size" and "margin" (if any) are a multiple of binning factor.
        See nabu:!208
        After that, all subregion lengths of _build_tasks_chunked() should be multiple of the binning factor,
        because "delta_z" itself was checked to be a multiple in DatasetValidator._handle_binning()
        """
        bin_z = self.process_config.binning_z
        if bin_z <= 1:
            return
        self.chunk_size -= self.chunk_size % bin_z
        if self._margin_v > 0 and (self._margin_v % bin_z) > 0:
            self._margin_v += bin_z - (self._margin_v % bin_z)  # i.e margin = ((margin % bin_z) + 1) * bin_z

    #
    # Tasks management
    #

    def _modify_processconfig_stage_1(self):
        # Modify the "process_config" object to dump sinograms
        proc = self.process_config
        self._old_steps_to_save = proc.steps_to_save.copy()
        if "sinogram" in proc.steps_to_save:
            return
        proc._configure_save_steps(self._old_steps_to_save + ["sinogram"])

    def _undo_modify_processconfig_stage_1(self):
        self.process_config.steps_to_save = self._old_steps_to_save
        if "sinogram" not in self._old_steps_to_save:
            self.process_config.dump_sinogram = False

    def _modify_processconfig_stage_2(self):
        # Modify the "process_config" object to resume from sinograms
        proc = self.process_config
        self._old_resume_from = proc.resume_from_step
        self._old_proc_steps = proc.processing_steps.copy()
        self._old_proc_options = proc.processing_options.copy()
        proc._configure_resume(resume_from="sinogram")

    def _undo_modify_processconfig_stage_2(self):
        self.process_config.resume_from_step = self._old_resume_from
        self.process_config.processing_steps = self._old_proc_steps
        self.process_config.processing_options = self._old_proc_options

    def _build_tasks_grouped(self):
        tasks = []
        segments = subdivide_into_overlapping_segment(self.n_angles, self.group_size, 0)
        for segment in segments:
            _, start_angle, end_angle, _ = segment
            z_min = max(self.z_min - self._margin_v, 0)
            z_max = min(self.z_max + self._margin_v, self.n_z)
            sub_region = ((start_angle, end_angle), (z_min, z_max), (0, self.chunk_shape[-1]))
            tasks.append({"sub_region": sub_region, "margin": (self.z_min - z_min, z_max - self.z_max)})
        self.tasks = tasks
        # Build tasks for stage 2 (sinogram processing + reconstruction)
        segments = subdivide_into_overlapping_segment(self.delta_z, self.chunk_size_sinos, 0)
        self._sino_tasks = []
        for segment in segments:
            _, start_z, end_z, _ = segment
            z_min = self.z_min + start_z
            z_max = min(self.z_min + end_z, self.n_z)
            sub_region = ((0, self.n_angles), (z_min, z_max), (0, self.chunk_shape[-1]))
            task = {"sub_region": sub_region, "margin": None}
            if self.process_config.dataset_info.flip_frame_ud:
                task["output_start_z"] = (
                    self.z_max - 1 - task["sub_region"][1][0] - self._get_task_vertical_margin(task)[0]
                )
            self._sino_tasks.append(task)

    def _build_tasks_chunked(self):
        margin_v = self._margin_v
        if self.chunk_size >= self.delta_z and self.z_min == 0 and self.z_max == self.n_z:
            # In this case we can do everything in a single stage, without margin
            self.tasks = [
                {
                    "sub_region": ((0, self.n_angles), (self.z_min, self.z_max), (0, self.chunk_shape[-1])),
                    "margin": None,
                }
            ]
            return
        if self.chunk_size - (2 * margin_v) >= self.delta_z:
            # In this case we can do everything in a single stage
            n_slices = self.delta_z
            (margin_up, margin_down) = (min(margin_v, self.z_min), min(margin_v, self.n_z - self.z_max))
            self.tasks = [
                {
                    "sub_region": (
                        (0, self.n_angles),
                        (self.z_min - margin_up, self.z_max + margin_down),
                        (0, self.chunk_shape[-1]),
                    ),
                    "margin": ((margin_up, margin_down), (0, 0)),
                }
            ]
            return
        # In this case there are at least two stages
        n_slices = self.chunk_size - (2 * margin_v)
        n_stages = ceil(self.delta_z / n_slices)
        self.tasks = []
        curr_z_min = self.z_min
        curr_z_max = self.z_min + n_slices
        for i in range(n_stages):
            margin_up = min(margin_v, curr_z_min)
            margin_down = min(margin_v, max(self.n_z - curr_z_max, 0))
            if curr_z_max + margin_down >= self.z_max:
                curr_z_max -= curr_z_max - (self.z_max + 0)
                margin_down = min(margin_v, max(self.n_z - 1 - curr_z_max, 0))
            self.tasks.append(
                {
                    "sub_region": (
                        (0, self.n_angles),
                        (int(curr_z_min - margin_up), int(curr_z_max + margin_down)),
                        (0, self.chunk_shape[-1]),
                    ),
                    "margin": ((margin_up, margin_down), (0, 0)),
                }
            )
            if curr_z_max == self.z_max:
                # No need for further tasks
                break
            curr_z_min += n_slices
            curr_z_max += n_slices

    def _build_tasks(self):
        if self._pipeline_mode == "grouped":
            self._build_tasks_grouped()
        else:
            self._ensure_good_chunk_size_and_margin()
            self._build_tasks_chunked()
        if self.process_config.dataset_info.flip_frame_ud:
            for task in self.tasks:
                # when applying up-down flip we will have a decreasing z index. Indices will be in [start_z, ..., end_z[
                task["output_start_z"] = (
                    self.z_max - 1 - task["sub_region"][1][0] - self._get_task_vertical_margin(task)[0]
                )

    def _get_task_vertical_margin(self, task):
        margin = task.get("margin", None)
        if margin is None:
            return (0, 0)
        if self._pipeline_mode == "grouped":
            # margin is in the form (up, down)
            # TODO this is not consistent
            return margin
        else:
            # margin is in the form ((up, down), (left, right))
            return margin[0]

    def _print_tasks_chunked(self):
        for task in self.tasks:
            margin_up, margin_down = task["margin"][0]
            s_u, s_d = task["sub_region"][1]
            print(
                "Top Margin: [%04d, %04d[  |  Slices: [%04d, %04d[  |  Bottom Margin: [%04d, %04d["
                % (s_u, s_u + margin_up, s_u + margin_up, s_d - margin_down, s_d - margin_down, s_d)
            )

    def _print_tasks(self):
        for task in self.tasks:
            margin_up, margin_down = task["margin"][0]
            s_u, s_d = task["sub_region"][1]
            print(
                "Top Margin: [%04d, %04d[  |  Slices: [%04d, %04d[  |  Bottom Margin: [%04d, %04d["  # pylint: disable=E1307
                % (s_u, s_u + margin_up, s_u + margin_up, s_d - margin_down, s_d - margin_down, s_d)
            )

    def _get_chunk_length(self, task):
        if self._pipeline_mode == "helical":
            (start_z, end_z) = task["sub_region"]
            return end_z - start_z

        else:
            (start_angle, end_angle), (start_z, end_z), _ = task["sub_region"]

            if self._pipeline_mode == "chunked":
                return end_z - start_z
            else:
                return end_angle - start_angle

    def _give_progress_info(self, task):
        self.logger.info("Processing sub-volume %s" % (str(task["sub_region"][:-1])))

    #
    # Reconstruction
    #

    def _reconstruct_chunked(self, tasks=None):
        self.results = {}
        self._histograms = {}
        self._data_dumps = {}
        tasks = tasks or self.tasks

        pipeline_kwargs = {}
        if self.backend == "cuda":
            pipeline_kwargs["cuda_options"] = self.cuda_options

        # Handle frames vertical flip
        if self.process_config.dataset_info.flip_frame_ud:
            self.logger.warning("Using 'flip_ud' setting. Reconstructed volume will be flipped upside down!")
            pipeline_kwargs["extra_options"] = {"write_in_reverse_order": True}

        # subdivide tasks into workers
        workers_tasks = split_list(tasks, self.n_workers)

        _log_options = None
        if isinstance(self.logger, Logger):
            _log_options = {"level": self.logger.level, "logfile": self.logger.logfile}

        if self._use_multiprocessing:
            tasks_processor_class = TaskProcessor
        else:
            tasks_processor_class = FakeTaskProcessor

        self.tasks_processor = tasks_processor_class(
            NannyPipeline,
            n_processes=self.n_workers,
            worker_init_args=(self.process_config,),
            worker_init_kwargs={
                "backend": self.backend,
                "pipeline_mode": self._pipeline_mode,
                "logging_options": _log_options,
            }
            | pipeline_kwargs,
        )

        def _result_callback(future):
            result = future.result()
            if "error" in result:
                return
            task_key = result["sub_region"]
            self.results[task_key] = result["result"]
            if result["histogram"] is not None:
                self._histograms[task_key] = result["histogram"]
            if len(result["data_dumps"]) > 0:
                self._data_dumps[task_key] = result["data_dumps"]

        self.tasks_processor.start_workers()
        try:
            # Submit tasks
            future_results = {}
            for worker_id in range(self.n_workers):
                current_worker_tasks = workers_tasks[worker_id]
                for task_id, task in enumerate(current_worker_tasks):
                    fut = self.tasks_processor.submit(worker_id, task)
                    fut.add_done_callback(_result_callback)
                    future_results[task_id] = fut
            self.tasks_processor.start_work()
            # Wait for all results. See also the above callback '_result_callback'
            while not (all(f.done() for f in future_results.values())):
                sleep(1)

            # Handle errors that could happen in workers sub-processes.
            # Note that future.result() is no longer blocking because of the above wait.
            # We raise a RuntimeError here, because it cannot be raised within the callback.
            for task_id, future in future_results.items():
                if "error" in future.result():
                    msg = f"Something went wrong with task {task_id} ! Aborting reconstruction"
                    self.logger.fatal(msg)
                    raise RuntimeError(msg)

        finally:
            # Terminate workers
            self.tasks_processor.stop_workers()

    def _reconstruct_grouped(self):
        self.results = {}
        self._data_dumps = {}

        # Stage 1: radios processing
        self._modify_processconfig_stage_1()

        self._reconstruct_chunked()
        self.merge_data_dumps(axis=0)

        # self._destroy_pipeline()
        self.logger.info("End of first stage of processing. Will now process sinograms saved on disk")
        self._undo_modify_processconfig_stage_1()

        # Stage 2: sinograms processing and reconstruction
        if "reconstruction" not in self.process_config.processing_steps:
            return
        self._modify_processconfig_stage_2()
        self._pipeline_mode = "chunked"
        self._reconstruct_chunked(tasks=self._sino_tasks)
        self._pipeline_mode = "grouped"
        self._undo_modify_processconfig_stage_2()

    def reconstruct(self):
        if self._pipeline_mode == "chunked":
            self._reconstruct_chunked()
        else:
            self._reconstruct_grouped()

    #
    # Writing data
    #

    def get_relative_files(self, files=None):
        out_cfg = self.process_config.nabu_config["output"]
        if files is None:
            files = list(self.results.values())
        if all(f is None for f in files):
            return []
        try:
            files.sort(key=variable_idxlen_sort)
        except:
            self.logger.error(
                "Lexical ordering failed, falling back to default sort - it will fail for more than 10k projections"
            )
            files.sort()
        local_files = [join(out_cfg["file_prefix"], basename(fname)) for fname in files]
        return local_files

    def _get_reconstruction_metadata(self, partial_volumes_files=None):
        metadata = {
            "nabu_config": self.process_config.nabu_config,
            "processing_options": self.process_config.processing_options,
        }
        if self._reconstruction_output_format_is_hdf5 and partial_volumes_files is not None:
            metadata[self._process_name + "_stages"] = {
                str(k): v for k, v in zip(self.results.keys(), partial_volumes_files)
            }
        if not (self._reconstruction_output_format_is_hdf5):
            metadata["process_info"] = {
                "process_name": "reconstruction",
                "processing_index": 0,
                "nabu_version": nabu_version,
            }
        return metadata

    def merge_hdf5_reconstructions(
        self,
        output_file=None,
        prefix=None,
        files=None,
        process_name=None,
        axis=0,
        merge_histograms=True,
        output_dir=None,
    ):
        """
        Merge existing hdf5 files by creating a HDF5 virtual dataset.

        Parameters
        ----------
        output_file: str, optional
            Output file name. If not given, the file prefix in section "output"
            of nabu config will be taken.
        """
        out_cfg = self.process_config.nabu_config["output"]
        out_dir = output_dir or out_cfg["location"]
        prefix = prefix or ""
        # Prevent issue when out_dir is empty, which happens only if dataset/location is a relative path.
        # TODO this should be prevented earlier
        if out_dir is None or len(out_dir.strip()) == 0:
            out_dir = dirname(dirname(self.results[first_generator_item(self.results.keys())]))
        #
        if output_file is None:
            output_file = join(out_dir, prefix + out_cfg["file_prefix"]) + ".hdf5"
        if isfile(output_file):
            msg = str("File %s already exists" % output_file)
            if out_cfg["overwrite_results"]:
                msg += ". Overwriting as requested in configuration file"
                self.logger.warning(msg)
            else:
                msg += ". Set overwrite_results to True in [output] to overwrite existing files."
                self.logger.fatal(msg)
                raise ValueError(msg)

        local_files = files
        if local_files is None:
            local_files = self.get_relative_files()
        if local_files == []:
            self.logger.error("No files to merge")
            return
        entry = getattr(self.process_config.dataset_info.dataset_scanner, "entry", "entry")
        process_name = process_name or self._process_name
        h5_path = join(entry, *[process_name, "results", "data"])
        #
        self.logger.info("Merging %ss to %s" % (process_name, output_file))
        #
        # When dumping to disk an intermediate step taking place before cropping the radios,
        # 'start_z' and 'end_z' found in nabu config have to be augmented with margin_z.
        # (these values are checked in ProcessConfig._configure_resume())
        #
        patched_start_end_z = False
        if (
            self._margin_v > 0
            and process_name != "reconstruction"
            and self.process_config.is_before_radios_cropping(process_name)
            and "reconstruction" in self.process_config.processing_steps
        ):
            user_rec_config = self.process_config.processing_options["reconstruction"]
            patched_start_end_z = True
            old_start_z = user_rec_config["start_z"]
            old_end_z = user_rec_config["end_z"]
            user_rec_config["start_z"] = max(0, old_start_z - self._margin_v)
            user_rec_config["end_z"] = min(self.n_z, old_end_z + self._margin_v)
        #
        merge_hdf5_files(
            local_files,
            h5_path,
            output_file,
            process_name,
            output_entry=entry,
            output_filemode="a",
            processing_index=0,
            config=self._get_reconstruction_metadata(local_files),
            base_dir=out_dir,
            axis=axis,
            overwrite=out_cfg["overwrite_results"],
        )
        if merge_histograms:
            self.merge_histograms(output_file=output_file)
        if patched_start_end_z:
            user_rec_config["start_z"] = old_start_z
            user_rec_config["end_z"] = old_end_z
        return output_file

    def merge_histograms(self, output_file=None, force_merge=False):
        """
        Merge the partial histograms
        """
        if not (self._do_histograms):
            return
        if self._histogram_merged and not (force_merge):
            return
        self.logger.info("Merging histograms")

        masterfile_entry = getattr(self.process_config.dataset_info.dataset_scanner, "entry", "entry")
        masterfile_process_name = "histogram"  # TODO don't hardcode process name
        output_entry = masterfile_entry

        out_cfg = self.process_config.nabu_config["output"]
        if output_file is None:
            output_file = (
                join(dirname(first_generator_item(self._histograms.values())), out_cfg["file_prefix"] + "_histogram")
                + ".hdf5"
            )
        local_files = self.get_relative_files(files=list(self._histograms.values()))
        #
        h5_path = join(masterfile_entry, *[masterfile_process_name, "results", "data"])
        #

        try:
            files = sorted(self._histograms.values(), key=variable_idxlen_sort)
        except:
            self.logger.error(
                "Lexical ordering of histogram failed, falling back to default sort - it will fail for more than 10k projections"
            )
            files = sorted(self._histograms.values())
        data_urls = []
        for fname in files:
            url = DataUrl(file_path=fname, data_path=h5_path, data_slice=None, scheme="silx")
            data_urls.append(url)
        histograms = []
        for data_url in data_urls:
            h2D = get_data(data_url)
            histograms.append((h2D[0], add_last_bin(h2D[1])))
        histograms_merger = PartialHistogram(method="fixed_bins_number", num_bins=histograms[0][0].size)
        merged_hist = histograms_merger.merge_histograms(histograms)

        rec_region = self.process_config.rec_region
        # Not sure if we should account for binning here.
        # (note that "rec_region" does not account for binning).
        # Anyway the histogram has little use when using binning
        volume_shape = (
            rec_region["end_z"] - rec_region["start_z"] + 1,
            rec_region["end_y"] - rec_region["start_y"] + 1,
            rec_region["end_x"] - rec_region["start_x"] + 1,
        )
        writer = NXProcessWriter(output_file, entry=output_entry, filemode="a", overwrite=True)
        writer.write(
            hist_as_2Darray(merged_hist),
            "histogram",
            processing_index=1,
            config={
                "files": local_files,
                "bins": self.process_config.nabu_config["postproc"]["histogram_bins"],
                "volume_shape": volume_shape,
            },
            is_frames_stack=False,
            direct_access=False,
        )
        self._histogram_merged = True

    def merge_data_dumps(self, axis=1):
        # Collect in a dict where keys are step names (instead of task keys)
        dumps = {}
        for task_key, data_dumps in self._data_dumps.items():  # noqa: PERF102
            for step_name, fname in data_dumps.items():
                fname = join(basename(dirname(fname)), basename(fname))
                if step_name not in dumps:
                    dumps[step_name] = [fname]
                else:
                    dumps[step_name].append(fname)
        # Merge HDF5 files
        for step_name, files in dumps.items():
            dump_file = self.process_config.get_save_steps_file(step_name=step_name)
            self.merge_hdf5_reconstructions(
                output_file=dump_file,
                output_dir=dirname(dump_file),
                files=files,
                process_name=step_name,
                axis=axis,
                merge_histograms=False,
            )

    def write_metadata_file(self):
        metadata = self._get_reconstruction_metadata()
        save_options = self.process_config.processing_options["save"]
        # Perhaps there is more elegant
        metadata_writer = VolumeSingleFrameBase(
            url=DataUrl(file_path=save_options["location"], data_path="/"),
            volume_basename=save_options["file_prefix"],
            overwrite=True,
            metadata=metadata,
        )
        metadata_writer.save_metadata()

    def finalize_files_saving(self):
        """
        Last step to save data. This will do several things:
          - Merge data dumps (which are always HDF5 files): create a master file for all data-dump sub-volumes
          - Merge HDF5 reconstruction (if output format is HDF5)
          - Create a "metadata file" (if output format is not HDF5)
          - Merge histograms (if output format is not HDF5)
        """
        self.merge_data_dumps()
        if "reconstruction" not in self.process_config.processing_steps:
            return
        if self._reconstruction_output_format_is_hdf5:
            self.merge_hdf5_reconstructions()
        else:
            self.merge_histograms()
            self.write_metadata_file()



from ..utils import MissingComponentError, deprecation_warning, warnings
from ..processing.processing_base import ProcessingBase
from .utils import __has_cupy__, dtype_to_ctype

if __has_cupy__:
    import cupy

    cuda = cupy.cuda
    cuarray = cupy.ndarray
    from ..cuda.kernel import CudaKernel
else:
    cuarray = MissingComponentError("cupy")
    dtype_to_ctype = MissingComponentError("cupy")


# NB: we must detach from a context before creating another context
class CudaProcessing(ProcessingBase):
    array_class = cuarray if __has_cupy__ else None
    dtype_to_ctype = dtype_to_ctype

    def __init__(self, device_id=None, ctx=None, stream=None, cleanup_at_exit=None):
        """
        Initialie a CudaProcessing instance.

        CudaProcessing is a base class for all CUDA-based processings.
        This class provides utilities for context/device management, and
        arrays allocation.

        Parameters
        ----------
        device_id: int, optional
            ID of the cuda device to use (those of the `nvidia-smi` command).
            Ignored if 'stream' is provided
        ctx: Unused parameter
            Deprecated, not used anymore.
        stream: cupy stream, optional
            Cuda stream. If not provided, will use the default stream
        cleanup_at_exit: bool, optional
            Deprecated, not used anymore
        """
        super().__init__()
        # COMPAT.
        if ctx is not None:
            deprecation_warning("Using 'ctx' kwarg is deprecated")
        if cleanup_at_exit is not None:
            deprecation_warning("Using 'cleanup_at_exit' kwarg is deprecated")
        #
        if stream is not None:
            stream.use()
        elif device_id is not None:
            cuda.Device(device_id).use()
        self.stream = stream  # TODO get default stream ?
        self.device_id = cuda.runtime.getDevice()
        self.device = cuda.Device(self.device_id)
        self.device_name = cuda.runtime.getDeviceProperties(self.device_id)["name"].decode()
        try:
            cuda.driver.ctxGetDevice()
        except cuda.driver.CUDADriverError:
            warnings.warn(f"No context found on cuda device {self.device_id}, creating one", RuntimeWarning)
            cuda.Device(self.device_id).use()

    def _allocate_array_mem(self, shape, dtype):
        return cupy.zeros(shape, dtype)

    def kernel(self, kernel_name, filename=None, src=None, automation_params=None, **build_kwargs):
        return CudaKernel(  # pylint: disable=E0606
            kernel_name,
            filename=filename,
            src=src,
            automation_params=automation_params,
            **build_kwargs,
        )



import numpy as np

from ..utils import updiv, get_cuda_srcfile
from ..cuda.utils import create_texture
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from .filtering_cuda import CudaSinoFilter
from .sinogram_cuda import CudaSinoMult
from .fbp_base import BackprojectorBase, ClipCircle


class CudaBackprojector(BackprojectorBase):
    backend = "cuda"
    kernel_filename = "backproj.cu"
    backend_processing_class = CudaProcessing
    SinoFilterClass = CudaSinoFilter
    SinoMultClass = CudaSinoMult

    def _check_textures_availability(self):
        self._use_textures = self.extra_options.get("use_textures", True)

    def _get_kernel_options(self):
        super()._get_kernel_options()
        self._kernel_options.update(
            {
                "file_name": get_cuda_srcfile(self.kernel_filename),
                "texture_name": "tex_projections",
            }
        )

    def _prepare_kernel_args(self):
        super()._prepare_kernel_args()
        self.kern_proj_kwargs.update(
            {
                "shared_size": self._kernel_options["shared_size"],
            }
        )
        if self._use_textures:
            self.kern_proj_args[1] = self._tex_sino
        else:
            self._d_sino = self._processing.allocate_array("_d_sino", self.sino_shape)
            self.kern_proj_args[1] = self._d_sino

    def _prepare_textures(self):
        if self._use_textures:
            self._tex_sino, self._d_sino_cua = create_texture(
                self.sino_shape, np.float32, address_mode="border", filter_mode="linear", normalized_coords=False
            )

    def _compile_kernels(self):
        self._prepare_textures()
        self._prepare_kernel_args()
        if self._use_textures:
            self._kernel_options["sourcemodule_options"].append("-DUSE_TEXTURES")
        self.gpu_projector = CudaKernel(
            self._kernel_options["kernel_name"],
            filename=self._kernel_options["file_name"],
            options=tuple(self._kernel_options["sourcemodule_options"]),
            silent_compilation_warnings=True,  # textures and Cuda 11
        )
        if self.halftomo and self.rot_center < self.dwidth:
            self.sino_mult = CudaSinoMult(self.sino_shape, self.rot_center)  # , ctx=self._processing.ctx)
        if self.extra_options["clip_outer_circle"]:
            self._clip_circle = CudaClipCircle(
                self.slice_shape, outer_circle_value=self.extra_options.get("outer_circle_value", 0)
            )

    def _get_filter_init_extra_options(self):
        return {
            # "cuda_options": {
            #     "ctx": self._processing.ctx,
            # },
        }

    def _transfer_to_texture(self, sino, do_checks=True):
        if do_checks and not (sino.flags.c_contiguous):
            raise ValueError("Expected C-Contiguous array")
        if self._use_textures:
            self._d_sino_cua.copy_from(sino)  # TODO stream ?
        else:
            if id(self._d_sino) == id(sino):
                return
            self._d_sino[:] = sino[:]


# COMPAT.
Backprojector = CudaBackprojector


class CudaClipCircle(ClipCircle):

    def _init_finalize(self, ctx_or_queue=None):
        # TODO (?) use CudaProcessing with 'ctx_or_queue' ?
        self._clip_circle_kernel = CudaKernel(
            "clip_circle",
            filename=get_cuda_srcfile("clip_circle.cu"),
        )
        self._clip_circle_block = (32, 32)
        self._clip_circle_kwargs = {
            "block": self._clip_circle_block,
            "grid": tuple([updiv(sz, blk) for sz, blk in zip(self.slice_shape[::-1], self._clip_circle_block)]),
        }

    def clip_circle(self, d_image):
        self._clip_circle_kernel(d_image, *self._clip_circle_args, **self._clip_circle_kwargs)

    __call__ = clip_circle


class PolarBackprojector(Backprojector):
    """
    Cuda Backprojector with output in polar coordinates.
    """

    cuda_fname = "backproj_polar.cu"
    cuda_kernel_name = "backproj_polar"

    # patch parent method: force slice_shape to (n_angles, n_x)
    def _set_angles(self, angles, n_angles):
        Backprojector._set_angles(self, angles, n_angles)
        self.slice_shape = (self.n_angles, self.n_x)

    # patch parent method:
    def _set_slice_roi(self, slice_roi):
        if slice_roi is not None:
            raise ValueError("slice_roi is not supported with this class")
        Backprojector._set_slice_roi(self, slice_roi)

    # patch parent method: don't do the 4X compute-workload optimization for this kernel
    def _get_kernel_options(self):
        Backprojector._get_kernel_options(self)
        block = self._kernel_options["block"]
        self._kernel_options["grid"] = (updiv(self.n_x, block[0]), updiv(self.n_y, block[1]))

    # patch parent method: update kernel args
    def _compile_kernels(self):
        n_y = self.n_y
        self.n_y = self.n_angles
        Backprojector._compile_kernels(self)
        self.n_y = n_y



import pytest
import numpy as np
from nabu.testutils import get_data
from nabu.processing.histogram import PartialHistogram
from nabu.cuda.utils import __has_cupy__

if __has_cupy__:
    from nabu.processing.histogram_cuda import CudaPartialHistogram


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("mri_rec_astra.npz")["data"]
    cls.data /= 10
    cls.data[:100] *= 10
    cls.data_part_1 = cls.data[:100]
    cls.data_part_2 = cls.data[100:]
    cls.data0 = cls.data.ravel()
    cls.bin_tol = 1e-5 * (cls.data0.max() - cls.data0.min())
    cls.hist_rtol = 1.5e-3
    if __has_cupy__:
        ...
    yield
    if __has_cupy__:
        ...


def add_nans_in_data(data_part1, data_part2, data_ref):
    data_part1.ravel()[0] = np.nan
    data_part2.ravel()[-1] = np.nan
    data_ref.ravel()[0] = np.nan
    data_ref.ravel()[-1] = np.nan


@pytest.mark.usefixtures("bootstrap")
class TestPartialHistogram:
    def compare_histograms(self, hist1, hist2):
        errmax_bins = np.max(np.abs(hist1[1] - hist2[1]))
        assert errmax_bins < self.bin_tol
        errmax_hist = np.max(np.abs(hist1[0] - hist2[0]) / hist2[0].max())
        assert errmax_hist / hist2[0].max() < self.hist_rtol

    def test_fixed_nbins(self):
        partial_hist = PartialHistogram(method="fixed_bins_number", num_bins=1e6)
        hist1 = partial_hist.compute_histogram(self.data_part_1.ravel())
        hist2 = partial_hist.compute_histogram(self.data_part_2.ravel())
        hist = partial_hist.merge_histograms([hist1, hist2])
        ref = np.histogram(self.data0, bins=partial_hist.num_bins)
        self.compare_histograms(hist, ref)

    def test_nans(self):
        partial_hist = PartialHistogram(method="fixed_bins_number", num_bins=1e6)
        data_part1 = self.data_part_1.copy()
        data_part2 = self.data_part_2.copy()
        data_ref = self.data0.copy()
        add_nans_in_data(data_part1, data_part2, data_ref)

        hist1 = partial_hist.compute_histogram(data_part1.ravel())
        hist2 = partial_hist.compute_histogram(data_part2.ravel())
        hist = partial_hist.merge_histograms([hist1, hist2])
        ref = np.histogram(data_ref, bins=partial_hist.num_bins, range=(np.nanmin(data_ref), np.nanmax(data_ref)))
        self.compare_histograms(hist, ref)

    @pytest.mark.skipif(not (__has_cupy__), reason="Need cuda/cupy for this test")
    def test_fixed_nbins_cuda(self):
        partial_hist = CudaPartialHistogram(method="fixed_bins_number", num_bins=1e6)
        data_part_1 = partial_hist.cuda_processing.to_device("data_part_1", np.tile(self.data_part_1, (1, 1, 1)))
        data_part_2 = partial_hist.cuda_processing.to_device("data_part_2", np.tile(self.data_part_2, (1, 1, 1)))
        hist1 = partial_hist.compute_histogram(data_part_1)
        hist2 = partial_hist.compute_histogram(data_part_2)
        hist = partial_hist.merge_histograms([hist1, hist2])
        ref = np.histogram(self.data0, bins=partial_hist.num_bins)
        self.compare_histograms(hist, ref)

    @pytest.mark.skipif(not (__has_cupy__), reason="Need cuda/cupy for this test")
    def test_nans_cuda(self):
        partial_hist = CudaPartialHistogram(method="fixed_bins_number", num_bins=1e6)
        data_part1 = self.data_part_1.copy()
        data_part2 = self.data_part_2.copy()
        data_ref = self.data0.copy()
        add_nans_in_data(data_part1, data_part2, data_ref)
        d_data_part1 = partial_hist.cuda_processing.to_device("data_part1", data_part1)
        d_data_part2 = partial_hist.cuda_processing.to_device("data_part2", data_part2)

        hist1 = partial_hist.compute_histogram(d_data_part1.reshape((1,) + d_data_part1.shape))
        hist2 = partial_hist.compute_histogram(d_data_part2.reshape((1,) + d_data_part2.shape))
        hist = partial_hist.merge_histograms([hist1, hist2])
        ref = np.histogram(data_ref, bins=partial_hist.num_bins, range=(np.nanmin(data_ref), np.nanmax(data_ref)))
        self.compare_histograms(hist, ref)



import os
import logging
from enum import Enum
from multiprocessing.pool import ThreadPool
from warnings import warn
import numpy as np
from ..pipeline.config_validators import convert_to_bool
from ..utils import check_supported, get_num_threads, is_scalar

try:
    from skimage.registration import phase_cross_correlation
except ImportError:
    phase_cross_correlation = None
from .translation import DetectorTranslationAlongBeam as RadiosShiftEstimator, estimate_shifts

plt = None
disable_matplotlib = convert_to_bool(os.environ.get("NABU_DISABLE_MATPLOTLIB", "0"))[0]
if not (disable_matplotlib):
    try:
        import matplotlib.pyplot as plt

        __have_matplotlib__ = True
    except ImportError:
        logging.getLogger(__name__).warning("Matplotlib not available or disabled. Plotting disabled")


class PairType(Enum):
    OPPOSITE = 1
    RETURN = 0


class MotionEstimation:
    """
    A class for estimating rigid sample motion during acquisition.
    The motion is estimated by
       1. Measuring the translations between radios (either opposite radios for 360-degrees scan ; or "reference radios")
       2. Fitting these measured translations with a displacement model.

    For (1), there are various available functions to estimate the shift between projections.
    The workhorse is phase cross-correlation, but this class allows to use other functions.
    For (2), you can pick several displacement model. The default one is a low-degree polynomial.

    Once the displacement model is computed, you have the displacement in sample reference frame,
    and you can "project" these displacements in the detector reference frame.
    Having the displacements converted as detector shifts allows you to
      (a) assess the fit between displacement model, and measured detector shifts
      (b) get a list movements to correct during reconstruction, vertical and/or horizontal.
          In nabu pipeline, this is handled by "translation_movements_file" in [reconstruction] section.
    """

    _shifts_estimators_default_kwargs = {
        "phase_cross_correlation": {"upsample_factor": 10, "overlap_ratio": 0.3},
        "DetectorTranslationAlongBeam": {"peak_fit_radius": 10, "padding_mode": "edge"},
        "estimate_shifts": {},
    }

    def __init__(
        self,
        projs_stack1,
        projs_stack2,
        angles1_rad,
        angles2_rad,
        indices1=None,
        indices2=None,
        n_projs_tot=None,
        n_calc_threads=None,
        reference="begin",
        shifts_estimator="phase_cross_correlation",
        shifts_estimator_kwargs=None,
    ):
        """
        Parameters
        ----------
        projs_stack1: numpy.ndarray
            Stack of projections, with shape (n_projs, n_rows, n_cols).
            It has to be of the same shape as 'projs_stack2'.
            Projection number 'i' in 'projs_stack1' will be compared with projection number 'i' in 'projs_stack2'
        projs_stack2: numpy.ndarray
            Stack of projections, with shape (n_projs, n_rows, n_cols).
            It has to be of the same shape as 'projs_stack1'.
            Projection number 'i' in 'projs_stack1' will be compared with projection number 'i' in 'projs_stack2'
        angles1_rad: numpy.ndarray or list of float
            Angles (in radians) of each projection, i.e, projection number 'i' in 'projs_stack1' was acquired at angle theta=angles1_rad[i]
        angles2_rad: numpy.ndarray or list of float
            Angles (in radians) of each projection, i.e, projection number 'i' in 'projs_stack2' was acquired at angle theta=angles2_rad[i]
        indices1: numpy.ndarray or list of int, optional
            Indices corresponding to each projection in 'projs_stack1'.
            It is used to calculate the curvilinear coordinate for the fit.
        indices2: numpy.ndarray or list of int, optional
            Indices corresponding to each projection in 'projs_stack2'.
            It is used to calculate the curvilinear coordinate for the fit.
            It is mostly important if projections in 'projs_stack2' are "return projections" (see Notes below for what this means)
        n_calc_threads: int, optional
            Number of threads to use for calculating phase cross correlation on pairs of images.
            Default is to use all available threads.

        Notes
        -----
        "Return projections" is the name of extra projections that might be acquired at the end of the scan.
        For example, for a 180-degrees scan with rotation angles [0, ..., 179.8], extra projections can be saved
        at angles [180, 90, 0]  (the rotation stage rewinds to its original angular position).
        These extra projection serve to check whether the sample/stage moved during the scan.
        In this case, angles2_rad = np.radians([180, 90, 0]).

        This class works by fitting the measured displacements (in detector space) with a model.
        This model uses a "normalized coordinate" built upon projection angles & indices.
        Each pair of radios (projs_stack1[i], projs_stack2[i]) has angles (angles1_rad[i], angles2_rad[i])
        and normalized coordinates (a1[i], a2[i])
        where a1 = normalize_coordinates(angles1_rad) and a2 = normalize_coordinates(angles2_rad)
        """
        self._set_projs_pairs(projs_stack1, projs_stack2, angles1_rad, angles2_rad)
        self._setup_normalized_coordinates(indices1, indices2, reference, n_projs_tot)
        self._configure_shifts_estimator(shifts_estimator, shifts_estimator_kwargs)

        self.e_theta = np.array([np.cos(self.angles1), -np.sin(self.angles1)]).T
        self._n_threads = n_calc_threads or max(1, get_num_threads() // 2)

        # Default value before fit
        self.shifts_vu = None
        self._coeffs = None
        self._coeffs_v = None
        self._deg_xy = None
        self._deg_z = None

    def _set_projs_pairs(self, projs_stack1, projs_stack2, angles1, angles2):
        if projs_stack1.shape != projs_stack2.shape:
            raise ValueError(
                f"'projs_stack1' and 'projs_stack2' must have the same shape - have {projs_stack1.shape} and {projs_stack2.shape}"
            )
        if len(angles1) != len(angles2):
            raise ValueError("'angles1' and 'angles2' must have the same size")
        if len(angles1) != projs_stack1.shape[0] or len(angles2) != projs_stack2.shape[0]:
            raise ValueError(
                "There must be as many values in (angles1, angles2) as there are projections in (projs_stack1, projs_stack2"
            )
        self.projs_stack1 = projs_stack1
        self.projs_stack2 = projs_stack2
        self.angles1 = np.array(angles1)
        self.angles2 = np.array(angles2)
        self.n_pairs = self.angles1.size
        # Now determine if the angular spacing between projs_stack1[i] and projs_stack1[i]
        self.pair_types = []
        gaps = []
        for i, (a1, a2) in enumerate(zip(self.angles1, self.angles2)):
            gap = np.mod(np.abs(a2 - a1), 2 * np.pi)
            if np.isclose(gap, 0, atol=1e-1):
                # same angle: probably "return projection"
                self.pair_types.append(PairType.RETURN)
            elif np.isclose(gap, np.pi, atol=1e-1):
                # opposed by 180 degrees
                self.pair_types.append(PairType.OPPOSITE)
            else:
                raise ValueError(
                    f"Projections pair number {i} are spaced by {np.degrees(gap)} degrees - don't know what to do with them"
                )
            gaps.append(gap)
        if np.std([pt.value for pt in self.pair_types]) > 0:
            raise NotImplementedError(
                "Mixing pairs (projection, opposite_projection) and (projection, return_projection) is not supported"
            )
        self.is_return = any([pt == PairType.RETURN for pt in self.pair_types])
        self.gaps = np.array(gaps)
        self._pair_types = np.array([pt.value for pt in self.pair_types])

    def _setup_normalized_coordinates(self, indices1, indices2, reference, n_projs_tot):
        self.indices1 = np.array(indices1) if indices1 is not None else None
        self.indices2 = np.array(indices2) if indices2 is not None else None
        self.n_projs_tot = n_projs_tot
        if n_projs_tot is None and indices1 is not None:
            self.n_projs_tot = np.max(indices1)  # best guess
        self.reference = reference
        # TODO find a better way
        distance_to_pi = abs(self.angles2.max() - np.pi)
        distance_to_twopi = abs(self.angles2.max() - 2 * np.pi)
        self._angle_max = 2 * np.pi if distance_to_twopi < distance_to_pi else np.pi
        self.a1 = self.normalize_coordinates(self.angles1, part=1)
        self.a2 = self.normalize_coordinates(self.angles2, part=2)
        self.a_all = np.concatenate([self.a1, self.a2])

    def normalize_coordinates(self, angles, part=1):
        """
        Get the "curvilinear coordinates" that are used (instead of projection angles in radians or degrees) for fit.
        These coordinates depend on:
           - how we normalize (wrt total number of angles, or wrt angle max)
           - the reference projection (start or end)

        Parameters
        ----------
        angles: array
            Array with projection angles
        part: int, optional
            Which part of the scan the provided angles belong to.
            Using "part=1" (resp. part=2) means that these angles correspond to "angles1_rad" (resp. angles2_rad)

        """
        # Currently this will work for
        #   - 360° scan with only pairs of opposite projs
        #   - 180° scan + pairs of return projections only  (we normalize only the angles of regular projs)
        # TODO consider other normalizations
        if part not in [1, 2]:
            raise ValueError("Expected 'part' to be either 1 or 2")

        def _normalize_with_reference(a):
            if self.reference == "end":
                return 1 - a
            return a

        # "Outward projections", always normalize with angle max
        if part == 1:  # noqa: SIM114
            return _normalize_with_reference(angles / self._angle_max)

        # Opposite projections (360° scan), normalize also with angle max
        elif not (self.is_return):
            return _normalize_with_reference(angles / self._angle_max)

        # Now, here, we are in the case "180° scan with return projections"
        #   - The regular projections have coordinate [0, ..., 1] (or almost 1)
        #   - The subsequent return projections have coordinates > 1: [1.01, ...]
        # We assume that "stack1" is all outward projs, "stack2" is all return projs (see NotImplementedError in _set_projs_pairs)
        n_outward_projs = self.angles1.size
        n_return_projs = self.angles2.size
        if self.indices2 is not None and self.n_projs_tot is not None:
            return _normalize_with_reference(self.indices2 / self.n_projs_tot)
        else:
            # This case is tricky, we should probably throw an error
            a_all = np.arange(n_outward_projs + n_return_projs) / n_outward_projs
            return _normalize_with_reference(a_all[-n_outward_projs:])

    def _configure_shifts_estimator(self, shifts_estimator, shifts_estimator_kwargs):
        check_supported(shifts_estimator, list(self._shifts_estimators_default_kwargs.keys()), "shifts estimator")
        shifts_estimator_kwargs = shifts_estimator_kwargs or {}
        self.shifts_estimator = shifts_estimator
        if self.shifts_estimator == "phase_cross_correlation" and phase_cross_correlation is None:
            warn(
                "shift estimator was set to 'phase_cross_correlation' but it requires scikit-image which is not available. Falling back to 'DetectorTranslationAlongBeam'",
                Warning,
            )
            self.shifts_estimator = "DetectorTranslationAlongBeam"
            shifts_estimator_kwargs.pop("overlap_ratio", None)  # Don't use this kwarg for DetectorTranslationAlongBeam
        self._shifts_estimator_kwargs = self._shifts_estimators_default_kwargs[self.shifts_estimator].copy()
        self._shifts_estimator_kwargs.update(shifts_estimator_kwargs)

    def _find_shifts(self, img1, img2):
        if self.shifts_estimator == "estimate_shifts":
            # estimate_shifts recovers the shifts in scipy convention (scipy.ndimage.shift),
            # but there is a sign difference wrt scikit-image.
            return -estimate_shifts(img1, img2, **self._shifts_estimator_kwargs)
        elif self.shifts_estimator == "phase_cross_correlation" and phase_cross_correlation is not None:
            return phase_cross_correlation(img1, img2, **self._shifts_estimator_kwargs)[0]
        else:
            return RadiosShiftEstimator().find_shift(np.stack([img1, img2]), [0, 1], **self._shifts_estimator_kwargs)

    def compute_detector_shifts(self):
        """This function computes the shifts between two images of all pairs."""

        def _comp_shift(i):
            img2 = self.projs_stack2[i]
            if self.pair_types[i] == PairType.OPPOSITE:
                img2 = img2[:, ::-1]
            return self._find_shifts(self.projs_stack1[i], img2)

        with ThreadPool(self._n_threads) as tp:
            shifts = tp.map(_comp_shift, range(self.n_pairs))

        self.shifts_vu = np.array(shifts)

    def _compute_detector_shifts_if_needed(self, recalculate_shifts):
        if recalculate_shifts or (self.shifts_vu is None):
            self.compute_detector_shifts()

    def get_model_matrix(self, do_cor_estimation=True, degree=1):
        """
        Compute the model matrix for horizontal components (x, y)
        For degree 1:
        M = [cos(theta) * (a^+ - a) ;  -sin(theta) * (a^+ - a) ; 2 ]
        This matrix needs three ingredients:
          - The angles "theta" for which pairs of radios are compared. We take the "angles1_rad" provided at this class instanciation.
          - The corresponding normalized coordinate "a" : a = normalize_coordinates(angles1_rad)
          - Normalized coordinate "a_plus" of each second radio in a pair.

        In other words, the pair of radios (projs_stack1[i], projs_stack2[i]) have angles (angles1_rad[i], angles2_rad[i])
        and normalized coordinates (a[i], a_plus[i])

        The resulting matrix strongly depends on how the angles are ordered/normalized.
        """
        angles = self.angles1
        cos_theta = np.cos(angles)
        sin_theta = np.sin(angles)
        ap = self.a2
        a = self.a1

        M = np.zeros((self.n_pairs, 2 * degree + int(do_cor_estimation)), dtype=np.float64)
        i_col = 0
        for d in range(degree, 0, -1):  # eg. [2, 1]. No 0 !
            columns = np.vstack([(ap**d - a**d) * cos_theta, -(ap**d - a**d) * sin_theta]).T
            M[:, i_col : i_col + 2] = columns
            i_col += 2
        if do_cor_estimation:
            M[:, -1] = 2
        return M

    def _get_vdm_matrix(self, a, degree):
        vdm_mat = np.stack([a**d for d in range(degree, 0, -1)], axis=1)
        return vdm_mat

    def apply_fit_horiz(self, angles=None, angles_normalized=None):
        """
        Apply the fitted parameters to get the sample displacement in (x, y)

        Parameters
        -----------
        angles: array, optional
            Angles the fit is applied onto.
        angles_normalized: array, optional
            Normalized angles the fit is applied onto. If provided, takes precedence over 'angles' (see notes below)

        Returns
        -------
        txy: array
            Array of shape (n_provided_angles, 2) where txy[:, 0] is the motion x-component, and txy[:, 1] is the motion y-component

        Notes
        ------
        The fit is assumed to have been done beforehand on a series of detector shifts measurements.
        Once the fit is done, coefficients are extracted and stored by this class instance.
        The parameter 'angles' provided to this function is normalized before applying the fit.
        For degree 2, applying the fit is roughly: t_x = alpha_x * a^2 + beta_x * a
        where 'a' is the normalized angle coordinate, and (alpha_x, beta_x) the coefficients extracted from the fit.
        """
        if self._coeffs is None:
            raise RuntimeError("Need to do estimate_horizontal_motion() first")
        if angles is None and angles_normalized is None:
            raise ValueError("Need to provide either 'angles' or 'angles_normalized'")
        if angles_normalized is None:
            angles_normalized = self.normalize_coordinates(angles)

        # See get_model_matrix()
        end = -1 if self._coeffs.size % 2 == 1 else None
        coeffs_x = self._coeffs[:end:2]
        coeffs_y = self._coeffs[1:end:2]

        vdm_mat = self._get_vdm_matrix(angles_normalized, self._deg_xy)
        return np.stack([vdm_mat.dot(coeffs_x), vdm_mat.dot(coeffs_y)], axis=1)

    def apply_fit_vertic(self, angles=None, angles_normalized=None):
        if self._coeffs_v is None:
            raise RuntimeError("Need to do estimate_vertical_motion() first")
        if angles is None and angles_normalized is None:
            raise ValueError("Need to provide either 'angles' or 'angles_normalized'")
        if angles_normalized is None:
            angles_normalized = self.normalize_coordinates(angles)
        vdm_mat = self._get_vdm_matrix(angles_normalized, self._deg_z)
        coeffs_z = self._coeffs_v
        return vdm_mat.dot(coeffs_z)

    def estimate_horizontal_motion(self, degree=1, cor=None, recalculate_shifts=False):
        """
        Estimation of the horizontal motion component.

        Parameters
        -----------
        degree: int (default=1).
            Degree of the polynomial model of the motion in the horizontal plane,
            for both x and y components.
        cor: float, optional
            Center of rotation, relative to the middle of the image.
            If None (default), it will be estimated along with the horizontal movement components.
            If provided (scalar value), use this value and estimate only the horizontal movement components.
        recalculate_shifts: bool, optional
            Whether to re-calculate detector shifts (usually with phase cross-correlation) if already calculated
        """
        do_cor_estimation = cor is None

        # Get "Delta u" estimated from pairs of projections
        self._compute_detector_shifts_if_needed(recalculate_shifts)

        # Build model matrix
        M = self.get_model_matrix(do_cor_estimation=do_cor_estimation, degree=degree)

        # Build parameters vector
        b = self.shifts_vu[:, 1] if do_cor_estimation else self.shifts_vu[:, 1] - 2 * cor * self._pair_types

        # Least-square fit, i.e  coeffs = pinv(M) * b
        self._coeffs = np.linalg.lstsq(M, b, rcond=None)[0]
        self._deg_xy = degree

        # Evaluate coefficients on current angles
        txy = self.apply_fit_horiz(angles_normalized=self.a_all)
        c = self._coeffs[-1] if do_cor_estimation else cor

        return txy, c

    estimate_cor_and_horizontal_motion = estimate_horizontal_motion

    def estimate_vertical_motion(self, degree=1, recalculate_shifts=False):
        """Estimation of the motion vertical component."""

        # Get "Delta v" estimated from pairs of projections
        self._compute_detector_shifts_if_needed(recalculate_shifts)

        # Model matrix is simple here: v = t_z  (parallel geometry)
        mat = np.zeros([self.n_pairs, degree])
        for d in range(degree, 0, -1):
            mat[:, degree - d] = self.a2**d - self.a1**d
        self._coeffs_v = np.linalg.lstsq(mat, self.shifts_vu[:, 0], rcond=None)[0]
        self._deg_z = degree

        tz = self.apply_fit_vertic(angles_normalized=self.a_all)
        return tz

    def convert_sample_motion_to_detector_shifts(self, t_xy, t_z, angles, cor=0):
        """
        Convert vectors of motion (t_x, t_y, t_z), from the sample domain,
        to vectors of motion (t_u, t_v) in the detector domain

        Parameters
        ----------
        t_xy: numpy.ndarray
            Sample horizontal shifts with shape (n_angles, 2).
            The first (resp. second) column are the x-shifts (resp. y-shifts)
        t_z: numpy.ndarray
            Sample vertical shifts, with the size n_angles
        angles: numpy.ndarray
            Rotation angle (in degree) corresponding to each component
        cor: float, optional
            Center of rotation
        """
        e_theta = np.array([np.cos(angles), -np.sin(angles)]).T
        dotp = np.sum(t_xy * e_theta, axis=1)
        shifts_u = dotp + cor
        shifts_v = t_z.copy()
        return np.stack([shifts_v, shifts_u], axis=1)

    def apply_fit_to_get_detector_displacements(self, cor=None):
        # This should be equal to:
        # txy2 = self.apply_fit_horiz(angles_normalized=self.a2)
        # txy1 = self.apply_fit_horiz(angles_normalized=self.a1)
        # shifts_u = np.sum((txy2 - txy1) * self.e_theta, axis=1) + 2 * cor * (1 - self.is_return)
        M = self.get_model_matrix(do_cor_estimation=(cor is None), degree=self._deg_xy)
        shifts_u = M.dot(self._coeffs) + 2 * (cor if cor is not None and not (self.is_return) else 0)

        tz2 = self.apply_fit_vertic(angles_normalized=self.a2)
        tz1 = self.apply_fit_vertic(angles_normalized=self.a1)
        shifts_v = tz2 - tz1

        shifts_vu = np.stack([shifts_v, shifts_u], axis=1)
        return shifts_vu

    def get_max_fit_error(self, cor=None):
        """
        Get the maximum error of the fit of displacement (in pixel units), projected in the detector domain ;
        i.e compare shifts_u (measured via phase cross correlation) to M.dot(coeffs) (model fit)
        """
        shifts_vu_modelled = self.apply_fit_to_get_detector_displacements(cor=cor)
        err_max_vu = np.max(np.abs(shifts_vu_modelled - self.shifts_vu), axis=0)
        return err_max_vu

    def plot_detector_shifts(self, cor=None):
        """
        Plot the detector shifts, i.e difference of u-movements between theta and theta^+.
        This can be used to compare the fit against measured shifts between pairs of projections.

        The sample movements were inferred from pairs of projections:
        projs1[i] at angles1[i], and projs2[i] at angles2[i]
        From these pairs of projections, a model of motion is built and we can generate:
          - The motion in sample domain (tx(theta), ty(theta), tz(theta)) for arbitrary theta
          - The motion in detector domain (t_u(theta), t_v(theta)) (for arbitrary theta) which is the parallel projection of the above

        What was used for the fit is t_u(theta^+) - t_u(theta).
        This function will plot this "difference" between t_u(theta^+) and t_u(theta).
        """
        estimated_shifts_vu = self.apply_fit_to_get_detector_displacements(cor=cor)
        gt_shifts_u = None
        gt_shifts_v = None

        fig, ax = plt.subplots(2)
        t_axis_vals = np.degrees(self.angles1)
        t_axis_label = "Angle (deg)"

        # Horizontal diff-movements on detector
        subplot = ax[0]
        subplot.plot(t_axis_vals, self.shifts_vu[:, 1], ".", label="Measured (%s)" % self.shifts_estimator)
        subplot.plot(t_axis_vals, estimated_shifts_vu[:, 1], ".-", label="Fit")
        if gt_shifts_u is not None:
            subplot.plot(self.angles1, gt_shifts_u, label="Ground Truth")
        subplot.set_xlabel(t_axis_label)
        subplot.set_ylabel("shift (pix)")
        subplot.legend()
        subplot.set_title("Horizontal shifts on the detector")

        # Vertical diff-movements on detector
        subplot = ax[1]
        subplot.plot(t_axis_vals, self.shifts_vu[:, 0], ".", label="Measured (%s)" % self.shifts_estimator)
        subplot.plot(t_axis_vals, estimated_shifts_vu[:, 0], ".-", label="Fit")
        if gt_shifts_v is not None:
            subplot.plot(t_axis_vals, gt_shifts_v, label="Ground truth")
        subplot.set_xlabel(t_axis_label)
        subplot.set_ylabel("shift (pix)")
        subplot.legend()
        subplot.set_title("Vertical shifts on the detector")

        plt.show()

    def plot_movements(self, cor=None, angles_rad=None, gt_xy=None, gt_z=None):
        """
        Plot the movements in the sample and detector domain, for a given vector of angles.
        Mind the difference with plot_sample_shifts(): in this case we plot proj(txy(theta))
        whereas plot_sample_shifts() plots proj(txy(theta^+)) - proj(txy(theta)),  which can be used to compare with measured shifts between projections.
        """
        if plt is None:
            raise ImportError("Need matplotlib")
        if self._deg_xy is None:
            raise RuntimeError("Need to estimate shifts first")

        if angles_rad is None:
            angles_rad = self.angles1
        angles_deg = np.degrees(angles_rad)

        txy = self.apply_fit_horiz(angles=angles_rad)
        tz = self.apply_fit_vertic(angles=angles_rad)
        estimated_shifts_vu = self.convert_sample_motion_to_detector_shifts(txy, tz, angles_rad, cor=cor or 0)

        gt_shifts_vu = None
        if gt_xy is not None:
            if gt_z is None or gt_z.shape[0] != gt_xy.shape[0] or gt_xy.shape[0] != len(angles_rad):
                raise ValueError("gt_xy has to be provided with gt_z, of the same size as 'angles_rad'")
            gt_shifts_vu = self.convert_sample_motion_to_detector_shifts(gt_xy, gt_z, angles_rad, cor=cor or 0)

        fig, ax = plt.subplots(2, 2)

        # Horizontal movements on detector
        subplot = ax[0, 0]
        subplot.plot(angles_deg, estimated_shifts_vu[:, 1], ".-", label="Fit")
        if gt_shifts_vu is not None:
            subplot.plot(angles_deg, gt_shifts_vu[:, 1], label="Ground Truth")
        subplot.set_xlabel("Angle (deg)")
        subplot.set_ylabel("shift (pix)")
        subplot.legend()
        subplot.set_title("Horizontal movements projected onto the detector")

        # Vertical movements on detector
        subplot = ax[1, 0]
        subplot.plot(angles_deg, estimated_shifts_vu[:, 0], ".-", label="Fit")
        if gt_shifts_vu is not None:
            subplot.plot(angles_deg, gt_shifts_vu[:, 0], label="Ground truth")
        subplot.set_xlabel("Angle (deg)")
        subplot.set_ylabel("shift (pix)")
        subplot.legend()
        subplot.set_title("Vertical movements projected onto the detector")

        # Horizontal movements in sample domain
        subplot = ax[0, 1]
        subplot.scatter(txy[:, 0], txy[:, 1], label="Fit", s=1)
        if gt_xy is not None:
            subplot.scatter(gt_xy[:, 0], gt_xy[:, 1], label="Ground Truth", s=1)
        subplot.set_xlabel("x (pix)")
        subplot.set_ylabel("y (pix)")
        subplot.legend()
        subplot.set_title("Sample movement in horizontal plane")

        # Vertical movements in sample domain
        subplot = ax[1, 1]
        subplot.plot(angles_deg, tz, ".-", label="Fit")
        if gt_z is not None:
            subplot.plot(angles_deg, gt_z, ".", label="Ground truth")
        subplot.set_xlabel("Angle (deg)")
        subplot.set_ylabel("z position (pix)")
        subplot.legend()
        subplot.set_title("Sample vertical movement")

        plt.show()
